/*---------------------------------------------------------------------------
:  This file contains utility procedures for operations with site control
:  information.
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
:  Header files
---------------------------------------------------------------------------*/

#  include  <stdio.h>
#  include  "comm.h"
#  include  "simtypes.h"
#  include  "extdata.h"

/*---------------------------------------------------------------------------
:  Control information initialization
---------------------------------------------------------------------------*/

initialize_ctl_info () 
{
   site_id_type site;
   site_state_type set_site_snum();

   actual_session_num = INITIAL_SNUM;
   failure_schedule = NULL_SUBFUNCTION;
   for (site = FIRST_SITE_ID ; site < num_sites ; site++)
      set_site_snum (site , INITIAL_SNUM);
}

/*---------------------------------------------------------------------------
:  ctl_info_io reads or writes status information for a site.  Its intended
:  use is for a site to write its status and the simulation manager to
:  read the status.
---------------------------------------------------------------------------*/

site_status_type  ctl_info_io ( function , site ) 

   int           function;
   site_id_type  site;
{
   char site_c[10] , open_mode;
   site_status_filename_type site_status_filename;
   FILE *site_status_file;
   site_status_type site_status;

   sprintf (site_c , "%d" , site);
   strcpy (site_status_filename , SITE_STATUS_ROOT_FILENAME);
   strcat (site_status_filename , site_c);

   open_mode = (function == WRITE_STATUS ? 'w' : 'r');
   site_status_file = fopen (site_status_filename , &open_mode);
   if (function == WRITE_STATUS)
      fprintf (site_status_file , "%d %c" , 
	       session_vector[site].snum , session_vector[site].state);
   else
      fscanf (site_status_file , "%d %c" , &site_status.session_num ,
	      &site_status.state);

   fclose (site_status_file);
   return (site_status);
}

/*---------------------------------------------------------------------------
:  get_site_state returns the current state of a specified site
---------------------------------------------------------------------------*/

site_state_type  get_site_state ( site )

   site_id_type  site;
{
   site_state_type  cur_state;

   cur_state = ((site >= FIRST_SITE_ID) && (site < num_sites) ?
		session_vector[site].state : NULL_STATE);
   return (cur_state);
}

/*---------------------------------------------------------------------------
:  get_old_site_state returns the old state of a specified site
---------------------------------------------------------------------------*/

site_state_type  get_old_site_state ( site )

   site_id_type  site;
{
   site_state_type old_state;

   old_state = ((site >= FIRST_SITE_ID) && (site < num_sites) ?
		old_session_vector[site].state : NULL_STATE);
   return (old_state);
}

/*---------------------------------------------------------------------------
:  set_site_state resets the state associated with a specified site
---------------------------------------------------------------------------*/

site_state_type  set_site_state ( site , new_state )
 
   site_id_type     site;
   site_state_type  new_state;
{
   if ((site >= FIRST_SITE_ID) && (site < num_sites)) {
      session_vector[site].state = new_state;
      if (site == MySID) {
         ctl_info_io (WRITE_STATUS , MySID);
         if (new_state == SITE_DOWN) failure_schedule = NULL_SUBFUNCTION;
      }
      return (new_state);
   }
   else
      return (NULL_STATE);
}

/*---------------------------------------------------------------------------
:  set_site_snum resets the session number associated with a specified site
---------------------------------------------------------------------------*/

site_state_type  set_site_snum ( site , new_snum )

   site_id_type  site;
   int           new_snum;
{
   if ((site >= FIRST_SITE_ID) && (site < num_sites)) {
      session_vector[site].snum = new_snum;
      session_vector[site].state = 
	 (new_snum >= INITIAL_SNUM ? SITE_UP : SITE_DOWN);
      failure_schedule = NULL_SUBFUNCTION;
      if (site == MySID) ctl_info_io (WRITE_STATUS , MySID);
      return (session_vector[site].state);
   }
   else
      return (NULL_STATE);
}

/*---------------------------------------------------------------------------
:  announce_site_failure sends a control message (type 2) to all UP sites to 
:  indicate that this site has detected failure of another site (or sites)
---------------------------------------------------------------------------*/

announce_site_failure ()
{
   int send_message();
   site_id_type site;
   msg_type *message; 

   message = (msg_type *) malloc (sizeof(msg_type));
   message->source = MySID;
   message->type = control;
   message->subtype = failure_announce;
   bcopy (session_vector , message->c.ctl.s_vector , sizeof(session_vector));

   for (site = FIRST_SITE_ID ; site < num_sites ; site++)
      if ((site != MySID) && (get_site_state (site) == SITE_UP)) {
	 message->dest = site;
	 send_message (message);
      }
   free (message);
   return (0);
}

/*-------------------------------------------------------------------------
 * send_recovery_response
 *-------------------------------------------------------------------------
 */
 send_recovery_response(destination)
	site_id_type	destination;
{
	msg_type 	*resp_mesg;

        /* log_site_info (DUMP , "recovery_response contents"); for debugging */
	resp_mesg = (msg_type *)malloc(sizeof(msg_type));
	resp_mesg->source  = MySID; 
	resp_mesg->dest    = destination;
	resp_mesg->type    = control;
	resp_mesg->subtype = recovery_response;
	bcopy(session_vector,resp_mesg->c.ctl.s_vector,
	             sizeof(session_vector_type));
	bcopy(fail_locks,resp_mesg->c.ctl.f_locks,
	             sizeof(fail_locks_type));
	send_message(resp_mesg);
	send_directory(destination) ; 
	free(resp_mesg);
}

/*------------------------------------------------------------------------
 * send_recovery_wait
 *------------------------------------------------------------------------
 */
 send_recovery_wait(destination)
	site_id_type	destination;
{
	msg_type	*resp_mesg;

	resp_mesg = (msg_type *)malloc(sizeof(msg_type));
	resp_mesg->source  = MySID; 
	resp_mesg->dest    = destination;
	resp_mesg->type    = control;
	resp_mesg->subtype = recovery_wait;
	send_message(resp_mesg);
	free(resp_mesg);
}

/*-------------------------------------------------------------------------
 * send_recovery_announce
 *-------------------------------------------------------------------------
 */
 send_recovery_announce(destination)
	site_id_type	destination;
{
	msg_type 	*resp_mesg;

	resp_mesg = (msg_type *)malloc(sizeof(msg_type));
	resp_mesg->source  = MySID; 
	resp_mesg->dest    = destination;
	resp_mesg->type    = control;
	resp_mesg->subtype = recovery_announce;
	send_message(resp_mesg);
	free(resp_mesg);
}

/*-------------------------------------------------------------------------
 * send_control_status
 *-------------------------------------------------------------------------
 */
 send_control_status(destination)
	site_id_type	destination;
{
	msg_type 	*resp_mesg;

	resp_mesg = (msg_type *)malloc(sizeof(msg_type));
	resp_mesg->source  = MySID; 
	resp_mesg->dest    = destination;
	resp_mesg->type    = control;
	resp_mesg->subtype = status;
	send_message(resp_mesg);
	free(resp_mesg);
}

/* a send_clear_fail_locks does not exist since it is complicated and
   must be done while processing a user xaction or after a copier xaction 
   (it is part of insure_current_database in typex.c) */
