/*---------------------------------------------------------------------------
:  This file contains utility procedures for database operations, 
:  including operations on the database itself and its fail-locks.
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
:  Flag to indicate if fail-lock processing should be performed during
:  committment of data-item values  (1 = yes , 0 = no)
---------------------------------------------------------------------------*/

#  define  FAIL_LOCKS_ENABLED  1

/*---------------------------------------------------------------------------
:  Header files
---------------------------------------------------------------------------*/

#  include  <stdio.h>
#  include  "comm.h"
#  include  "simtypes.h"
#  include  "extdata.h"
/*************/
#  include  "direct.h"
/*************/

/*---------------------------------------------------------------------------
:  Procedures to manage database data-items
:
:  initialize_database sets up data-item names and values.  The names used
:  here are integers equal to the index of the data-item.
---------------------------------------------------------------------------*/

initialize_database ()
{
   int  item_index;
   data_item_type cur_item;
   data_value_type item_value = INITIAL_ITEM_VALUE;

   for (item_index = 0 ; item_index < MAX_NUM_ITEMS ; item_index++) {
      cur_item = item_index;
      database[item_index].item = cur_item;
      data_item_mgr (WRITE_ITEM , database[item_index].item , &item_value);
      fail_locks_mgr 
	 (CLEAR_FOR_UP_SITES , NULL_SITE_ID , database[item_index].item);
   }
   return (0);
}

/*---------------------------------------------------------------------------
:  commit_db_values writes values into the database based on the contents
:  of an xact.update message
---------------------------------------------------------------------------*/

commit_db_values ( upd_msg )

   msg_type *upd_msg;
{
   int index;

   if (last_update == (msg_type *)NULL) 
      last_update = (msg_type *) malloc (sizeof (msg_type));
   bcopy (upd_msg , last_update , sizeof (msg_type));

   for (index = 0 ; index < upd_msg->c.db.num_items ; index++) {
/**********/
     if (directory_mgr(CHECK_FOR_COPY, MySID, upd_msg->c.db.item[index]) ==
		COPY_PRESENT )  
/**********/
          data_item_mgr (WRITE_ITEM , upd_msg->c.db.item[index] , 
		     &upd_msg->c.db.value[index]);

#     if FAIL_LOCKS_ENABLED == 1
      fail_locks_mgr (CLEAR_FOR_UP_SITES , NULL_SITE_ID , 
		      upd_msg->c.db.item[index]); 
      fail_locks_mgr (SET_FOR_DOWN_SITES , NULL_SITE_ID ,
		      upd_msg->c.db.item[index]);
#     endif
   }
   return (0);
}

/*---------------------------------------------------------------------------
:  data_item_mgr manages all accesses to the database
---------------------------------------------------------------------------*/

data_item_mgr ( function , arg_item , arg_value )

   db_action_type   function;
   data_item_type   arg_item;
   data_value_type  *arg_value;
{

/*---------------------------------------------------------------------------
:  Local variable definition
---------------------------------------------------------------------------*/

   int ret_index = ITEM_NOT_FOUND , item_index = 0;

/*---------------------------------------------------------------------------
:  Start of code - look up the data-item in the database
---------------------------------------------------------------------------*/

   while ((database[item_index].item != arg_item) && 
	  (item_index < MAX_NUM_ITEMS))  item_index++;

/*---------------------------------------------------------------------------
:  Continue processing according to function argument
---------------------------------------------------------------------------*/

   if (item_index < MAX_NUM_ITEMS) {
      ret_index = item_index;

      switch (function) {
	 case READ_ITEM :
	      *arg_value = database[item_index].value;
	      break;

         case WRITE_ITEM :
	      database[item_index].value = *arg_value;
              break;

         case INDEX_ITEM :
	      /* do nothing since ret_index set above */
	      break;

         default :
	      printf ("*** Illegal data-item function ***\n");
	      break;
      }
   }

   return (ret_index);
}

/*---------------------------------------------------------------------------
:  Procedure to manage database fail-locks
---------------------------------------------------------------------------*/

fail_locks_mgr ( function , site , item)

   int             function;
   site_id_type    site;
   data_item_type  item;
{

/*---------------------------------------------------------------------------
:  Local variable definition
---------------------------------------------------------------------------*/

   site_state_type get_site_state();
   site_id_type site_id;
   int status = FAIL_LOCK_ERROR , index , site_mask;

/*---------------------------------------------------------------------------
:  Start of code - look up the data-item and perform requested function
---------------------------------------------------------------------------*/

   if ((index = data_item_mgr (INDEX_ITEM , item , NULL_ITEM_VALUE)) !=
       ITEM_NOT_FOUND) {

      switch (function) {

/*---------------------------------------------------------------------------
:        To set a fail lock, create a bit mask with the bit set for the
:        site in question and OR this mask with the item fail locks 
---------------------------------------------------------------------------*/

	 case SET_FOR_SITE :
	      status = FAIL_LOCK_SET;
	      site_mask = 0x01 << site;  
	      fail_locks[index] = fail_locks[index] | site_mask;
	      break;

         case SET_FOR_DOWN_SITES :
	      status = FAIL_LOCK_SET;
	      for (site_id = FIRST_SITE_ID ; site_id < num_sites ; site_id++)
                 if (get_site_state (site_id) != SITE_UP) {
/**********  if the item exits on the failed site  */
     		    if (directory_mgr(CHECK_FOR_COPY, site_id, item) == 
				COPY_PRESENT /*&& directory_mgr( CHECK_FOR_COPY,
				MySID, item) == COPY_PRESENT*/)  { 
/**********/
		        site_mask = 0x01 << site_id;
		        fail_locks[index] = fail_locks[index] | site_mask;
		    } 
		 }
              break;

/*---------------------------------------------------------------------------
:        To clear a fail lock, create a bit mask with all bits set except
:        for the bit for the site in question and AND this mask with the 
:        item fail locks 
---------------------------------------------------------------------------*/

         case CLEAR_FOR_SITE :
	      status = FAIL_LOCK_CLEAR;
	      site_mask = ~ (0x01 << site);
              fail_locks[index] = fail_locks[index] & site_mask;
	      break;

         case CLEAR_FOR_UP_SITES :
              status = FAIL_LOCK_CLEAR;
	      for (site_id = FIRST_SITE_ID ; site_id < num_sites ; site_id++)
                 if (get_site_state(site_id) == SITE_UP) {
/**********/
     		    if (directory_mgr(CHECK_FOR_COPY, site_id, item) == 
					COPY_PRESENT )  { 
/**********/
		       site_mask = ~ (0x01 << site_id);
		       fail_locks[index] = fail_locks[index] & site_mask;
		    } 
		 }
	      break;

/*---------------------------------------------------------------------------
:        To check a fail lock, create a bit mask with the bit set for the
:        site in question and determine if the AND of the mask and the
:        fail locks produces a non-zero result
---------------------------------------------------------------------------*/

         case CHECK_FOR_SITE :
	      site_mask = 0x01 << site; 
	      status = ((fail_locks[index] & site_mask) == 0 ?
			FAIL_LOCK_CLEAR : FAIL_LOCK_SET);
	      break;

         default :
	      printf ("*** Illegal fail-locks function ***\n");
	      break;
      }
   }
   return (status);
}

/*---------------------------------------------------------------------------
:  Procedure to count fail-locks
---------------------------------------------------------------------------*/

count_fail_locks (site)

   site_id_type site;
{
   int num_fail_locks = 0 , index;

   for (index = 0 ; index < MAX_NUM_ITEMS ; index++)
      if (fail_locks_mgr (CHECK_FOR_SITE , site , index) == FAIL_LOCK_SET)
         num_fail_locks ++;
   
   return (num_fail_locks);
}
