/*---------------------------------------------------------------------------
:  Header file:   simtypes.h
:  Dependencies:  "comm.h" 
:
:  This file contains definitions and data type structures needed for the 
:  distributed database simulation.
---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
:  Define the number of sites to be supported (the maximum number of sites is
:  partially dependent on the number of sites the communication sw supports)
---------------------------------------------------------------------------*/

#define  MIN_NUM_SITES          2
#define  MAX_NUM_DESIRED_SITES  15

#if MAX_SITES > MAX_NUM_DESIRED_SITES
#   define  MAX_NUM_SITES  MAX_NUM_DESIRED_SITES 
#else
#   define  MAX_NUM_SITES  MAX_SITES
#endif

/*---------------------------------------------------------------------------
:  Define site ID formats
---------------------------------------------------------------------------*/

#define  FIRST_SITE_ID      0
#define  LAST_SITE_ID       MAX_NUM_SITES - 1
#define  SIMULATION_MGR_ID  MAX_NUM_SITES
#define  NULL_SITE_ID       -1
typedef  int  site_id_type;

/*---------------------------------------------------------------------------
:  Define the different states that a site can take on 
---------------------------------------------------------------------------*/

#define  SITE_UP              'U' 
#define  SITE_DOWN            'D' 
#define  SITE_RECOVERY_WAIT   'W' 
#define  SITE_STATE_UNKNOWN   '?' 
#define  SITE_TERMINATING     'T' 
#define  NULL_STATE           'N'
typedef  char  site_state_type;

/*---------------------------------------------------------------------------
:  Define data types for session number and the session vector
---------------------------------------------------------------------------*/

#define  INITIAL_SNUM  1
#define  NULL_SNUM    -1
typedef  int  actual_snum_type;

typedef  struct {
	 char  state;
	 int   snum;
         }  session_vector_record;

typedef  session_vector_record  session_vector_type [MAX_NUM_SITES];

/*---------------------------------------------------------------------------
:  Define the database operations and structure
---------------------------------------------------------------------------*/

#define  DB_UP_TO_DATE  1
#define  DB_OBSOLETE    2

#define  READ_ITEM   'R'
#define  WRITE_ITEM  'W'
#define  INDEX_ITEM  'I'
#define  NULL_ACTION 'N'
typedef  char  db_action_type;

#define  MAX_NUM_ITEMS       100 
#define  ITEM_NOT_FOUND       -1
#define  INITIAL_ITEM_VALUE  999 
#define  MIN_ITEM_VALUE        1
#define  MAX_ITEM_VALUE      500
#define  NULL_ITEM_VALUE      -1
typedef  short int  data_item_type;
typedef  short int  data_value_type;

typedef  struct {
	 data_item_type  item;
	 data_value_type value;
         }  db_record_type;

typedef  db_record_type  database_type [MAX_NUM_ITEMS];  

/*---------------------------------------------------------------------------
:  Define operations and structure of fail-locks
---------------------------------------------------------------------------*/

#define  SET_FOR_SITE        1
#define  SET_FOR_DOWN_SITES  2
#define  CLEAR_FOR_SITE      3
#define  CLEAR_FOR_UP_SITES  4
#define  CHECK_FOR_SITE      5

#define  FAIL_LOCK_SET       1
#define  FAIL_LOCK_CLEAR     2
#define  FAIL_LOCK_ERROR     3

typedef  short int  fail_locks_type [MAX_NUM_ITEMS]; 

/*
   Define structure for directories
*/

typedef struct {
	short int	direct;
	unsigned char	num_copies;
	unsigned char	min_copies;
	} directory_type [MAX_NUM_ITEMS];


/*---------------------------------------------------------------------------
:  Define information for the execution log and site status files to be kept
:  by each site
---------------------------------------------------------------------------*/

#define  EXEC_LOG_ROOT_FILENAME  "log."
typedef  char  exec_log_filename_type [10];
#define  NO_DUMP  1
#define  DUMP     2
#define  SUMMARY  3
#define  TRANS_MSG_SIZE  30

#define  SITE_STATUS_ROOT_FILENAME  "stat."
typedef  char  site_status_filename_type [10];
#define  READ_STATUS   1
#define  WRITE_STATUS  2
typedef  struct {
	 int              session_num; 
	 site_state_type  state;
         }  site_status_type;

/*---------------------------------------------------------------------------
:  Define the message types to be supported
:
:     control       : these messages are associated with site transitions
:     xact          : these messages are associated with db transactions
:     managing      : these messages are used to manage the simulation
---------------------------------------------------------------------------*/

enum  message_types  { control , xact , managing }; 

/*---------------------------------------------------------------------------
:  Define the message subtypes
:
:     control message subtypes:
:         recovery_announce : used by a site to announce that it is prepared
:                             to become operational (corresponds to type 1)
:         recovery_response : used by operational site to send session
:                             vector and fail locks to a recovering site
:         recovery_wait : sent by a recovering site to another recovering
:                         site if both sites must await the recovery of
:                         the last site to fail
:         failure_announce : sent by an operational site to other sites
:                            after detection of site failure (corresponds to
:                            type 2)
:         clear_fail_locks : sent to clear a fail lock after a copier xact
:	  directory_copy : sent to request a copy of the directories
:				from an operational site.
:	  directory_ack : sent in response to a directory_copy.
:				contains current copy of all directories.
:         status : sent by the last site down after all sites have failed
:                  to determine which sites are awaiting its recovery
:
:     xact message subtypes:
:         user : a database transaction initiated by a user
:         update : a transaction to write values from the user xact 
:                  to the database of another site
:         ack : acknowledge reception of "xact update"
:         commit : commit the transaction
:         commit_ack : acknowledge committment
:         copier : a transaction to read values from the database of
:                  another site
:         copier_update : a transaction to return requested copier values
:
:     managing message subtypes:
:         stop : cause a site to gracefully terminate after completion of 
:                the simulation
:         revive : cause a site that is down to start recovering
:         die : cause a site that is up to go into failure mode
:         dump : get information from a site
:         up : used in response to a control.status message to indicate
:	       that the site is up
:         failed : used by a site which is in failure mode to indicate to
:                  another site that it is not functioning (instead of
:                  having the operational site time out) 
:         xact_committed , xact_aborted : sent to the manager to indicate
:                                         the outcome of a xact.user
:	  allow_recovery : used to simulate the window of vunerability between 
:		the time that a site sends out a recovery_announce and the time 
:		it receives the recovery_response. When a site receives this 
:               message it will send the recovery_response.
---------------------------------------------------------------------------*/

enum  message_subtypes
      { recovery_announce , recovery_response ,   /* control subtypes */
	recovery_wait , failure_announce , directory_copy , directory_ack ,
	clear_fail_locks , status,                       

        user , update , direct_update , copier ,  /* xact subtypes */
	copier_update , ack , commit , commit_ack ,


	stop , revive , die , dump , up, failed , /* managing subtypes */
	xact_committed , xact_aborted , allow_recovery };

/*---------------------------------------------------------------------------
:  Messages are considered to have four control fields followed by the
:  message contents.  Note that the size of the message structure is one
:  less than the maximum message length to allow room for the string
:  terminator.
---------------------------------------------------------------------------*/

enum  message_types     tmp_msg_types;
enum  message_subtypes  tmp_msg_subtypes;

#define  MSG_CTL_FIELD_LEN  (sizeof (site_id_type) +  \
			     sizeof (site_id_type) +  \
			     sizeof (tmp_msg_types) + \
			     sizeof (tmp_msg_subtypes))

#define  MAX_CONTENTS  MAX_MSG_LEN - MSG_CTL_FIELD_LEN - 1

/*---------------------------------------------------------------------------
:  Those messages which are completely defined within the control fields
:  require no contents.
---------------------------------------------------------------------------*/

typedef  int  empty_contents_type [MAX_CONTENTS];

/*---------------------------------------------------------------------------
:  Messages for user transactions have this types of contents
---------------------------------------------------------------------------*/

#define  MAX_NUM_XACT_ACTIONS  25

typedef  struct {
	 db_action_type  action;
	 data_item_type  item;
	 data_value_type value;
         }  db_operation_type;

typedef  struct {
         int  num_actions;
	 db_operation_type  op [MAX_NUM_XACT_ACTIONS];
         }  user_xact_type;

/*---------------------------------------------------------------------------
:  Messages which must pass information about database items have
:  this type of contents.
---------------------------------------------------------------------------*/

typedef  struct {
	 int  num_pseudo_xacts;
	 int  num_copier_xacts;
         int  num_items;
	 data_item_type  item [MAX_NUM_ITEMS];
	 data_value_type value [MAX_NUM_ITEMS];
	 }  item_list_type;

/*---------------------------------------------------------------------------
:  Messages which must pass database control information have this type
:  of contents.
---------------------------------------------------------------------------*/

typedef  struct {
	 site_id_type	        subj_site;	/* subject of managing.tbd */
	 session_vector_type  	s_vector;
	 fail_locks_type      	f_locks;
	 }  control_info_type;


/*---------------------------------------------------------------------------
:  Messages of type managing which need ancillary information have this type
:  of contents.
---------------------------------------------------------------------------*/

#define  NULL_SUBFUNCTION     0
#define  FAIL_IMMEDIATELY     1
#define  FAIL_AFTER_UPD_RECV  2  
#define  FAIL_AFTER_UPD_SEND  3

typedef  int  man_subfunction_type;

/*---------------------------------------------------------------------------
:  Define the complete message structure
---------------------------------------------------------------------------*/

typedef  union  {
         empty_contents_type  empty;
         item_list_type       db;
         control_info_type    ctl;
	 directory_type       ctl_direct;
	 user_xact_type       ux;
	 man_subfunction_type subfunction;
         }  contents_type;

typedef  struct {
	 site_id_type            source;
	 site_id_type            dest;
	 enum  message_types     type; 
	 enum  message_subtypes  subtype;
	 contents_type           c;
	 }  msg_type;

/*---------------------------------------------------------------------------
:  Miscellaneous definitions
---------------------------------------------------------------------------*/

typedef  int  boolean;
#define  TRUE   1
#define  FALSE  0

/*---------------------------------------------------------------------------
:  End of simulation types
---------------------------------------------------------------------------*/
